@@ -16,8 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: table.c,v 1.15 2003/05/21 21:42:16 nalin Exp $"
-#include "../config.h"
+#include <config.h>
 #include <sys/types.h>
 #include <assert.h>
 #include <ctype.h>
@@ -31,42 +30,130 @@
 #include "table.h"
 
 /* Table info. */
-#define VTE_TABLE_MAX_LITERAL 160
+#define VTE_TABLE_MAX_LITERAL (128 + 32)
 #define _vte_table_map_literal(__c) \
 	(((__c) < (VTE_TABLE_MAX_LITERAL)) ? (__c) : 0)
 #define _vte_table_is_numeric(__c) \
-	((((__c) >= '0') && ((__c) <= '9')) || ((__c) == ';'))
-enum _vte_table_specials {
-	_vte_table_string = VTE_TABLE_MAX_LITERAL,
-	_vte_table_number,
-	_vte_table_max
-};
+	(((__c) >= '0') && ((__c) <= '9'))
+#define _vte_table_is_numeric_list(__c) \
+	((((__c) >= '0') && ((__c) <= '9')) || (__c) == ';')
+
 struct _vte_table {
+	struct _vte_matcher_impl impl;
 	GQuark resultq;
 	const char *result;
 	unsigned char *original;
 	gssize original_length;
 	int increment;
-	struct _vte_table *table[_vte_table_max];
+	struct _vte_table *table_string;
+	struct _vte_table *table_number;
+	struct _vte_table *table_number_list;
+	struct _vte_table **table;
 };
 
 /* Argument info. */
 enum _vte_table_argtype {
-	_vte_table_arg_number,
+	_vte_table_arg_number=0,
 	_vte_table_arg_string,
 	_vte_table_arg_char
 };
 struct _vte_table_arginfo {
-	enum _vte_table_argtype type;
 	const gunichar *start;
-	gssize length;
+	struct _vte_table_arginfo *next;
+	guint type:2;
+	guint length:30;
 };
+#define MAX_STACK 16
+struct _vte_table_arginfo_head {
+	guint stack_allocated;
+	struct _vte_table_arginfo *list;
+	struct _vte_table_arginfo stack[MAX_STACK];
+};
+
+static void
+_vte_table_arginfo_head_init(struct _vte_table_arginfo_head *head)
+{
+	head->list = NULL;
+	head->stack_allocated = 0;
+}
+static inline struct _vte_table_arginfo*
+_vte_table_arginfo_alloc(struct _vte_table_arginfo_head *head)
+{
+	struct _vte_table_arginfo *info;
+	if (G_LIKELY (head->stack_allocated < G_N_ELEMENTS(head->stack))) {
+		info = &head->stack[head->stack_allocated++];
+	} else {
+		info = g_slice_new (struct _vte_table_arginfo);
+	}
+	info->next = head->list;
+	head->list = info;
+	return info;
+}
+static void
+_vte_table_arginfo_head_revert(struct _vte_table_arginfo_head *head, struct _vte_table_arginfo *last)
+{
+	struct _vte_table_arginfo *info;
+	info = head->list;
+	head->list = last->next;
+	if (last >= &head->stack[0] &&
+			last < &head->stack[G_N_ELEMENTS(head->stack)]){
+		head->stack_allocated = last - &head->stack[0];
+	}
+	do {
+		struct _vte_table_arginfo *next = info->next;
+		if (info >= &head->stack[0] &&
+				info < &head->stack[G_N_ELEMENTS(head->stack)]){
+			break;
+		}
+		g_slice_free(struct _vte_table_arginfo, info);
+		if (info == last) {
+			break;
+		}
+		info = next;
+	}while (TRUE);
+}
+static struct _vte_table_arginfo *
+_vte_table_arginfo_head_reverse(struct _vte_table_arginfo_head *head)
+{
+	struct _vte_table_arginfo *prev = NULL;
+	while (head->list) {
+		struct _vte_table_arginfo *next = head->list->next;
+
+		head->list->next = prev;
+
+		prev = head->list;
+		head->list = next;
+	}
+	return prev;
+}
+static void
+_vte_table_arginfo_head_finalize(struct _vte_table_arginfo_head *head)
+{
+	struct _vte_table_arginfo *info, *next;
+	for (info = head->list; info != NULL; info = next) {
+		next = info->next;
+		if (info >= &head->stack[0] &&
+				info < &head->stack[G_N_ELEMENTS(head->stack)]){
+			continue;
+		}
+		g_slice_free(struct _vte_table_arginfo, info);
+	}
+}
 
 /* Create an empty, one-level table. */
 struct _vte_table *
 _vte_table_new(void)
 {
-	return g_malloc0(sizeof(struct _vte_table));
+	struct _vte_table * ret;
+	ret = g_slice_new0(struct _vte_table);
+	ret->impl.klass = &_vte_matcher_table;
+	return ret;
+}
+
+static struct _vte_table **
+_vte_table_literal_new(void)
+{
+	return g_new0(struct _vte_table *, VTE_TABLE_MAX_LITERAL);
 }
 
 /* Free a table. */
@@ -74,11 +161,22 @@ void
 _vte_table_free(struct _vte_table *table)
 {
 	unsigned int i;
-	for (i = 0; i < G_N_ELEMENTS(table->table); i++) {
-		if (table->table[i] != NULL) {
-			_vte_table_free(table->table[i]);
-			table->table[i] = NULL;
+	if (table->table != NULL) {
+		for (i = 0; i < VTE_TABLE_MAX_LITERAL; i++) {
+			if (table->table[i] != NULL) {
+				_vte_table_free(table->table[i]);
+			}
 		}
+		g_free(table->table);
+	}
+	if (table->table_string != NULL) {
+		_vte_table_free(table->table_string);
+	}
+	if (table->table_number != NULL) {
+		_vte_table_free(table->table_number);
+	}
+	if (table->table_number_list != NULL) {
+		_vte_table_free(table->table_number_list);
 	}
 	if (table->original_length == 0) {
 		g_assert(table->original == NULL);
@@ -86,11 +184,9 @@ _vte_table_free(struct _vte_table *table)
 		g_assert(table->original != NULL);
 	}
 	if (table->original != NULL) {
-		table->original_length = 0;
 		g_free(table->original);
-		table->original = NULL;
 	}
-	g_free(table);
+	g_slice_free(struct _vte_table, table);
 }
 
 /* Add a string to the tree with the given increment value. */
@@ -104,24 +200,27 @@ _vte_table_addi(struct _vte_table *table,
 	guint8 check;
 	struct _vte_table *subtable;
 
+	if (original_length == -1) {
+		original_length = strlen((char *) original);
+	}
+	if (length == -1) {
+		length = strlen(pattern);
+	}
+
 	/* If this is the terminal node, set the result. */
 	if (length == 0) {
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_PARSE)) {
-			if (table->result != NULL) {
-				g_warning("`%s' and `%s' are indistinguishable",
+		if (table->result != NULL)
+			_vte_debug_print (VTE_DEBUG_PARSE, 
+					  "`%s' and `%s' are indistinguishable.\n",
 					  table->result, result);
-			}
-		}
-#endif
+
 		table->resultq = g_quark_from_string(result);
 		table->result = g_quark_to_string(table->resultq);
 		if (table->original != NULL) {
 			g_free(table->original);
 		}
-		table->original = g_malloc(original_length);
+		table->original = g_memdup(original, original_length);
 		table->original_length = original_length;
-		memcpy(table->original, original, original_length);
 		table->increment = inc;
 		return;
 	}
@@ -139,13 +238,50 @@ _vte_table_addi(struct _vte_table *table,
 		/* Handle numeric parameters. */
 		if ((pattern[1] == 'd') ||
 		    (pattern[1] == '2') ||
-		    (pattern[1] == 'm')) {
+		    (pattern[1] == '3')) {
 			/* Create a new subtable. */
-			if (table->table[_vte_table_number] == NULL) {
+			if (table->table_number == NULL) {
 				subtable = _vte_table_new();
-				table->table[_vte_table_number] = subtable;
+				table->table_number = subtable;
 			} else {
-				subtable = table->table[_vte_table_number];
+				subtable = table->table_number;
+			}
+			/* Add the rest of the string to the subtable. */
+			_vte_table_addi(subtable, original, original_length,
+					pattern + 2, length - 2,
+					result, quark, inc);
+			return;
+		}
+
+		/* Handle variable-length parameters. */
+		if ((pattern[1] == 'm') ||
+		    (pattern[1] == 'M')) {
+			/* Build the "new" original using the initial portion
+			 * of the original string and what's left after this
+			 * specifier. */
+			if (pattern[1] == 'm') {
+				int initial;
+				GByteArray *b;
+
+				initial = original_length - length;
+				/* 0 args; we use 'M' to signal that zero is
+				 * not allowed.  */
+				b = g_byte_array_new();
+				g_byte_array_set_size(b, 0);
+				g_byte_array_append(b, original, initial);
+				g_byte_array_append(b, (const guint8*)pattern + 2, length - 2);
+				_vte_table_addi(table, b->data, b->len,
+						(const char *)b->data + initial,
+						b->len - initial,
+						result, quark, inc);
+				g_byte_array_free(b, TRUE);
+			}
+			/* Create a new subtable. */
+			if (table->table_number_list == NULL) {
+				subtable = _vte_table_new();
+				table->table_number_list = subtable;
+			} else {
+				subtable = table->table_number_list;
 			}
 			/* Add the rest of the string to the subtable. */
 			_vte_table_addi(subtable, original, original_length,
@@ -159,11 +295,11 @@ _vte_table_addi(struct _vte_table *table,
 			/* It must have a terminator. */
 			g_assert(length >= 3);
 			/* Create a new subtable. */
-			if (table->table[_vte_table_string] == NULL) {
+			if (table->table_string == NULL) {
 				subtable = _vte_table_new();
-				table->table[_vte_table_string] = subtable;
+				table->table_string = subtable;
 			} else {
-				subtable = table->table[_vte_table_string];
+				subtable = table->table_string;
 			}
 			/* Add the rest of the string to the subtable. */
 			_vte_table_addi(subtable, original, original_length,
@@ -175,6 +311,11 @@ _vte_table_addi(struct _vte_table *table,
 		/* Handle an escaped '%'. */
 		if (pattern[1] == '%') {
 			/* Create a new subtable. */
+			if (table->table == NULL) {
+				table->table = _vte_table_literal_new();
+				subtable = _vte_table_new();
+				table->table['%'] = subtable;
+			} else
 			if (table->table['%'] == NULL) {
 				subtable = _vte_table_new();
 				table->table['%'] = subtable;
@@ -196,6 +337,11 @@ _vte_table_addi(struct _vte_table *table,
 			 * character value. */
 			for (i = pattern[2]; i < VTE_TABLE_MAX_LITERAL; i++) {
 				/* Create a new subtable. */
+				if (table->table == NULL) {
+					table->table = _vte_table_literal_new();
+					subtable = _vte_table_new();
+					table->table[i] = subtable;
+				} else
 				if (table->table[i] == NULL) {
 					subtable = _vte_table_new();
 					table->table[i] = subtable;
@@ -209,6 +355,11 @@ _vte_table_addi(struct _vte_table *table,
 						result, quark, inc);
 			}
 			/* Also add a subtable for higher characters. */
+			if (table->table == NULL) {
+				table->table = _vte_table_literal_new();
+				subtable = _vte_table_new();
+				table->table[0] = subtable;
+			} else
 			if (table->table[0] == NULL) {
 				subtable = _vte_table_new();
 				table->table[0] = subtable;
@@ -226,12 +377,18 @@ _vte_table_addi(struct _vte_table *table,
 	/* A literal (or an unescaped '%', which is also a literal). */
 	check = (guint8) pattern[0];
 	g_assert(check < VTE_TABLE_MAX_LITERAL);
+	if (table->table == NULL) {
+		table->table = _vte_table_literal_new();
+		subtable = _vte_table_new();
+		table->table[check] = subtable;
+	} else
 	if (table->table[check] == NULL) {
 		subtable = _vte_table_new();
 		table->table[check] = subtable;
 	} else {
 		subtable = table->table[check];
 	}
+
 	/* Add the rest of the string to the subtable. */
 	_vte_table_addi(subtable, original, original_length,
 			pattern + 1, length - 1,
@@ -244,34 +401,19 @@ _vte_table_add(struct _vte_table *table,
 	       const char *pattern, gssize length,
 	       const char *result, GQuark quark)
 {
-	char *pattern_copy, *p;
-	pattern_copy = g_strndup(pattern, length);
-	/* Collapse as many numeric parameters as possible into '%m'. */
-	while ((p = strstr(pattern_copy, "%d")) != NULL) {
-		memcpy(p, "%m", 2);
-	}
-	while ((p = strstr(pattern_copy, "%3")) != NULL) {
-		memcpy(p, "%m", 2);
-	}
-	while ((p = strstr(pattern_copy, "%2")) != NULL) {
-		memcpy(p, "%m", 2);
-	}
-	while ((p = strstr(pattern_copy, "%m;%m")) != NULL) {
-		memmove(p, p + 3, length - 3 - (p - pattern_copy));
-		length -= 3;
-	}
-	_vte_table_addi(table, pattern_copy, length, pattern_copy, length,
+	_vte_table_addi(table,
+			(const unsigned char *) pattern, length,
+			pattern, length,
 			result, quark, 0);
-	g_free(pattern_copy);
 }
 
 /* Match a string in a subtree. */
 static const char *
 _vte_table_matchi(struct _vte_table *table,
-		  const gunichar *pattern, gssize length,
+		  const gunichar *candidate, gssize length,
 		  const char **res, const gunichar **consumed, GQuark *quark,
 		  unsigned char **original, gssize *original_length,
-		  GList **params)
+		  struct _vte_table_arginfo_head *params)
 {
 	int i = 0;
 	struct _vte_table *subtable = NULL;
@@ -279,7 +421,7 @@ _vte_table_matchi(struct _vte_table *table,
 
 	/* Check if this is a result node. */
 	if (table->result != NULL) {
-		*consumed = pattern;
+		*consumed = candidate;
 		*original = table->original;
 		*original_length = table->original_length;
 		*res = table->result;
@@ -289,108 +431,129 @@ _vte_table_matchi(struct _vte_table *table,
 
 	/* If we're out of data, but we still have children, return the empty
 	 * string. */
-	if ((length == 0) && (table != NULL)) {
-		*consumed = pattern;
+	if (G_UNLIKELY (length == 0)) {
+		*consumed = candidate;
 		return "";
 	}
 
 	/* Check if this node has a string disposition. */
-	if (table->table[_vte_table_string] != NULL) {
+	if (table->table_string != NULL) {
 		/* Iterate over all non-terminator values. */
-		subtable = table->table[_vte_table_string];
+		subtable = table->table_string;
 		for (i = 0; i < length; i++) {
-			if (subtable->table[_vte_table_map_literal(pattern[i])] != NULL) {
+			if ((subtable->table != NULL) &&
+			    (subtable->table[_vte_table_map_literal(candidate[i])] != NULL)) {
 				break;
 			}
 		}
 		/* Save the parameter info. */
-		arginfo = g_malloc(sizeof(struct _vte_table_arginfo));
+		arginfo = _vte_table_arginfo_alloc(params);
 		arginfo->type = _vte_table_arg_string;
-		arginfo->start = pattern;
+		arginfo->start = candidate;
 		arginfo->length = i;
-		*params = g_list_append(*params, arginfo);
 		/* Continue. */
-		return _vte_table_matchi(subtable, pattern + i, length - i,
+		return _vte_table_matchi(subtable, candidate + i, length - i,
 					 res, consumed, quark,
 					 original, original_length, params);
 	}
 
-	/* Check if this could be a number. */
-	if ((_vte_table_is_numeric(pattern[0])) &&
-	    (table->table[_vte_table_number] != NULL)) {
-		subtable = table->table[_vte_table_number];
-		/* Iterate over all numeric characters. */
-		for (i = 0; i < length; i++) {
-			if (!_vte_table_is_numeric(pattern[i])) {
+	/* Check if this could be a list. */
+	if ((_vte_table_is_numeric_list(candidate[0])) &&
+	    (table->table_number_list != NULL)) {
+		const char *local_result;
+
+		subtable = table->table_number_list;
+		/* Iterate over all numeric characters and ';'. */
+		for (i = 1; i < length; i++) {
+			if (!_vte_table_is_numeric_list(candidate[i])) {
 				break;
 			}
 		}
 		/* Save the parameter info. */
-		arginfo = g_malloc(sizeof(struct _vte_table_arginfo));
+		arginfo = _vte_table_arginfo_alloc(params);
 		arginfo->type = _vte_table_arg_number;
-		arginfo->start = pattern;
+		arginfo->start = candidate;
+		arginfo->length = i;
+
+		/* Try and continue. */
+		local_result = _vte_table_matchi(subtable,
+					 candidate + i, length - i,
+					 res, consumed, quark,
+					 original, original_length,
+					 params);
+		if (local_result != NULL) {
+			return local_result;
+		}
+		_vte_table_arginfo_head_revert (params, arginfo);
+
+		/* try again */
+	}
+
+	/* Check if this could be a number. */
+	if ((_vte_table_is_numeric(candidate[0])) &&
+	    (table->table_number != NULL)) {
+		subtable = table->table_number;
+		/* Iterate over all numeric characters. */
+		for (i = 1; i < length; i++) {
+			if (!_vte_table_is_numeric(candidate[i])) {
+				break;
+			}
+		}
+		/* Save the parameter info. */
+		arginfo = _vte_table_arginfo_alloc(params);
+		arginfo->type = _vte_table_arg_number;
+		arginfo->start = candidate;
 		arginfo->length = i;
-		*params = g_list_append(*params, arginfo);
 		/* Continue. */
-		return _vte_table_matchi(subtable, pattern + i, length - i,
+		return _vte_table_matchi(subtable, candidate + i, length - i,
 					 res, consumed, quark,
 					 original, original_length, params);
 	}
 
 	/* Check for an exact match. */
-	if (table->table[_vte_table_map_literal(pattern[0])] != NULL) {
-		subtable = table->table[_vte_table_map_literal(pattern[0])];
+	if ((table->table != NULL) &&
+	    (table->table[_vte_table_map_literal(candidate[0])] != NULL)) {
+		subtable = table->table[_vte_table_map_literal(candidate[0])];
 		/* Save the parameter info. */
-		arginfo = g_malloc(sizeof(struct _vte_table_arginfo));
+		arginfo = _vte_table_arginfo_alloc(params);
 		arginfo->type = _vte_table_arg_char;
-		arginfo->start = pattern;
+		arginfo->start = candidate;
 		arginfo->length = 1;
-		*params = g_list_append(*params, arginfo);
 		/* Continue. */
-		return _vte_table_matchi(subtable, pattern + 1, length - 1,
+		return _vte_table_matchi(subtable, candidate + 1, length - 1,
 					 res, consumed, quark,
 					 original, original_length, params);
 	}
 
 	/* If there's nothing else to do, then we can't go on.  Keep track of
 	 * where we are. */
-	*consumed = pattern;
+	*consumed = candidate;
 	return NULL;
 }
 
 static void
-_vte_table_extract_number(GValueArray **array,
-			  struct _vte_table_arginfo *arginfo, long increment)
+_vte_table_extract_numbers(GValueArray **array,
+			   struct _vte_table_arginfo *arginfo, long increment)
 {
 	GValue value = {0,};
-	GString *tmp;
-	char **vals;
-	int i;
+	gssize i;
 
-	tmp = g_string_new("");
-	for (i = 0; i < arginfo->length; i++) {
-		tmp = g_string_append_unichar(tmp, arginfo->start[i]);
-	}
-
-	vals = g_strsplit(tmp->str, ";", -1);
-
-	if (vals != NULL) {
-		g_value_init(&value, G_TYPE_LONG);
-
-		for (i = 0; vals[i] != NULL; i++) {
-			if (*array == NULL) {
-				*array = g_value_array_new(1);
-			}
-			g_value_set_long(&value, atol(vals[i]));
-			g_value_array_append(*array, &value);
+	g_value_init(&value, G_TYPE_LONG);
+	i = 0;
+	do {
+		long total = 0;
+		for (; i < arginfo->length && arginfo->start[i] != ';'; i++) {
+			gint v = g_unichar_digit_value (arginfo->start[i]);
+			total *= 10;
+			total += v == -1 ?  0 : v;
 		}
-
-		g_strfreev(vals);
-
-		g_value_unset(&value);
-	}
-
-	g_string_free(tmp, TRUE);
+		if (G_UNLIKELY (*array == NULL)) {
+			*array = g_value_array_new(1);
+		}
+		g_value_set_long(&value, total);
+		g_value_array_append(*array, &value);
+	} while (i++ < arginfo->length);
+	g_value_unset(&value);
 }
 
 static void
@@ -399,18 +562,17 @@ _vte_table_extract_string(GValueArray **array,
 {
 	GValue value = {0,};
 	gunichar *ptr;
-	int i;
+	guint i;
 
-	ptr = g_malloc(sizeof(gunichar) * (arginfo->length + 1));
-	memcpy(ptr, arginfo->start, (arginfo->length * sizeof(gunichar)));
+	ptr = g_new(gunichar, arginfo->length + 1);
 	for (i = 0; i < arginfo->length; i++) {
-		ptr[i] &= ~(VTE_ISO2022_ENCODED_WIDTH_MASK);
+		ptr[i] = arginfo->start[i] & ~VTE_ISO2022_ENCODED_WIDTH_MASK;
 	}
-	ptr[arginfo->length] = '\0';
+	ptr[i] = '\0';
 	g_value_init(&value, G_TYPE_POINTER);
 	g_value_set_pointer(&value, ptr);
 
-	if (*array == NULL) {
+	if (G_UNLIKELY (*array == NULL)) {
 		*array = g_value_array_new(1);
 	}
 	g_value_array_append(*array, &value);
@@ -426,7 +588,7 @@ _vte_table_extract_char(GValueArray **array,
 	g_value_init(&value, G_TYPE_LONG);
 	g_value_set_long(&value, *(arginfo->start) - increment);
 
-	if (*array == NULL) {
+	if (G_UNLIKELY (*array == NULL)) {
 		*array = g_value_array_new(1);
 	}
 	g_value_array_append(*array, &value);
@@ -436,86 +598,93 @@ _vte_table_extract_char(GValueArray **array,
 /* Check if a string matches something in the tree. */
 const char *
 _vte_table_match(struct _vte_table *table,
-		 const gunichar *pattern, gssize length,
+		 const gunichar *candidate, gssize length,
 		 const char **res, const gunichar **consumed,
 		 GQuark *quark, GValueArray **array)
 {
 	struct _vte_table *head;
-	const gunichar *dummy_consumed = NULL;
-	const char *dummy_res = NULL;
-	GQuark dummy_quark = 0;
-	GValueArray *dummy_array = NULL;
-	const char *ret = NULL;
-	unsigned char *original = NULL, *p = NULL;
+	const gunichar *dummy_consumed;
+	const char *dummy_res;
+	GQuark dummy_quark;
+	GValueArray *dummy_array;
+	const char *ret;
+	unsigned char *original, *p;
 	gssize original_length;
-	GList *params = NULL, *tmp;
 	long increment = 0;
 	int i;
+	struct _vte_table_arginfo_head params;
 	struct _vte_table_arginfo *arginfo;
 
 	/* Clean up extracted parameters. */
-	if (res == NULL) {
+	if (G_UNLIKELY (res == NULL)) {
 		res = &dummy_res;
 	}
 	*res = NULL;
-	if (consumed == NULL) {
+	if (G_UNLIKELY (consumed == NULL)) {
 		consumed = &dummy_consumed;
 	}
-	*consumed = pattern;
-	if (quark == NULL) {
+	*consumed = candidate;
+	if (G_UNLIKELY (quark == NULL)) {
 		quark = &dummy_quark;
 	}
 	*quark = 0;
-	if (array == NULL) {
+	if (G_UNLIKELY (array == NULL)) {
+		dummy_array = NULL;
 		array = &dummy_array;
 	}
-	*array = NULL;
 
 	/* Provide a fast path for the usual "not a sequence" cases. */
-	if (length == 0) {
-		return NULL;
-	}
-	if (pattern == NULL) {
+	if (G_LIKELY (length == 0 || candidate == NULL)) {
 		return NULL;
 	}
 
 	/* If there's no literal path, and no generic path, and the numeric
 	 * path isn't available, then it's not a sequence, either. */
-	if (table->table[_vte_table_map_literal(pattern[0])] == NULL) {
-		if (table->table[_vte_table_string] == NULL) {
-			if (!(_vte_table_is_numeric(pattern[0])) ||
-			    (table->table[_vte_table_number] == NULL)) {
-				/* No match. */
-				return NULL;
+	if (table->table == NULL ||
+	    table->table[_vte_table_map_literal(candidate[0])] == NULL) {
+		if (table->table_string == NULL) {
+			if (table->table_number == NULL ||
+					!_vte_table_is_numeric(candidate[0])){
+				if (table->table_number_list == NULL ||
+					!_vte_table_is_numeric_list(candidate[0])){
+					/* No match. */
+					return NULL;
+				}
 			}
 		}
 	}
 
 	/* Check for a literal match. */
-	for (i = 0, head = table; (i < length) && (head != NULL); i++) {
-		head = head->table[_vte_table_map_literal(pattern[i])];
+	for (i = 0, head = table; i < length && head != NULL; i++) {
+		if (head->table == NULL) {
+			head = NULL;
+		} else {
+			head = head->table[_vte_table_map_literal(candidate[i])];
+		}
 	}
-	if ((head != NULL) && (head->result != NULL)) {
+	if (head != NULL && head->result != NULL) {
 		/* Got a literal match. */
-		*consumed = pattern + i;
+		*consumed = candidate + i;
 		*res = head->result;
 		*quark = head->resultq;
 		return *res;
 	}
 
+	_vte_table_arginfo_head_init (&params);
+
 	/* Check for a pattern match. */
-	ret = _vte_table_matchi(table, pattern, length,
+	ret = _vte_table_matchi(table, candidate, length,
 				res, consumed, quark,
 				&original, &original_length,
 				&params);
 	*res = ret;
 
 	/* If we got a match, extract the parameters. */
-	if ((ret != NULL) && (strlen(ret) > 0) && (array != &dummy_array)) {
-		tmp = params;
+	if (ret != NULL && ret[0] != '\0' && array != &dummy_array) {
 		g_assert(original != NULL);
 		p = original;
-		while (p < original + original_length) {
+		arginfo = _vte_table_arginfo_head_reverse (&params);
+		do {
 			/* All of the interesting arguments begin with '%'. */
 			if (p[0] == '%') {
 				/* Handle an increment. */
@@ -525,59 +694,41 @@ _vte_table_match(struct _vte_table *table,
 					continue;
 				}
 				/* Handle an escaped '%'. */
-				if (p[1] == '%') {
-					tmp = g_list_next(tmp);
-					p += 2;
-					continue;
+				else if (p[1] == '%') {
 				}
 				/* Handle numeric parameters. */
-				if ((p[1] == 'd') ||
+				else if ((p[1] == 'd') ||
 				    (p[1] == '2') ||
-				    (p[1] == 'm')) {
-					arginfo = tmp->data;
-					_vte_table_extract_number(array,
-								  arginfo,
-								  increment);
-					tmp = g_list_next(tmp);
-					p += 2;
-					continue;
+				    (p[1] == '3') ||
+				    (p[1] == 'm') ||
+				    (p[1] == 'M')) {
+					_vte_table_extract_numbers(array,
+								   arginfo,
+								   increment);
+					p++;
 				}
 				/* Handle string parameters. */
-				if (p[1] == 's') {
-					arginfo = tmp->data;
+				else if (p[1] == 's') {
 					_vte_table_extract_string(array,
 								  arginfo);
-					tmp = g_list_next(tmp);
-					p += 2;
-					continue;
+					p++;
 				}
 				/* Handle a parameter character. */
-				if (p[1] == '+') {
-					arginfo = tmp->data;
+				else if (p[1] == '+') {
 					_vte_table_extract_char(array,
 								arginfo,
 								p[2]);
-					tmp = g_list_next(tmp);
-					p += 3;
-					continue;
+					p += 2;
+				} else {
+					g_assert_not_reached();
 				}
-				g_assert_not_reached();
-			} else {
-				/* Literal. */
-				tmp = g_list_next(tmp);
-				p++;
-				continue;
-			}
-		}
+			} /* else Literal. */
+			arginfo = arginfo->next;
+		} while (++p < original + original_length && arginfo);
 	}
 
 	/* Clean up extracted parameters. */
-	if (params != NULL) {
-		for (tmp = params; tmp != NULL; tmp = g_list_next(tmp)) {
-			g_free(tmp->data);
-		}
-		g_list_free(params);
-	}
+	_vte_table_arginfo_head_finalize (&params);
 
 	return ret;
 }
@@ -592,13 +743,13 @@ _vte_table_printi(struct _vte_table *table, const char *lead, int *count)
 
 	/* Result? */
 	if (table->result != NULL) {
-		fprintf(stderr, "%s = `%s'(%d)\n", lead,
+		g_printerr("%s = `%s'(%d)\n", lead,
 			table->result, table->increment);
 	}
 
 	/* Literal? */
 	for (i = 1; i < VTE_TABLE_MAX_LITERAL; i++) {
-		if (table->table[i] != NULL) {
+		if ((table->table != NULL) && (table->table[i] != NULL)) {
 			if (i < 32) {
 				newlead = g_strdup_printf("%s^%c", lead,
 							  i + 64);
@@ -611,17 +762,17 @@ _vte_table_printi(struct _vte_table *table, const char *lead, int *count)
 	}
 
 	/* String? */
-	if (table->table[_vte_table_string] != NULL) {
+	if (table->table_string != NULL) {
 		newlead = g_strdup_printf("%s{string}", lead);
-		_vte_table_printi(table->table[_vte_table_string],
+		_vte_table_printi(table->table_string,
 				  newlead, count);
 		g_free(newlead);
 	}
 
 	/* Number(+)? */
-	if (table->table[_vte_table_number] != NULL) {
+	if (table->table_number != NULL) {
 		newlead = g_strdup_printf("%s{number}", lead);
-		_vte_table_printi(table->table[_vte_table_number],
+		_vte_table_printi(table->table_number,
 				  newlead, count);
 		g_free(newlead);
 	}
@@ -633,7 +784,7 @@ _vte_table_print(struct _vte_table *table)
 {
 	int count = 0;
 	_vte_table_printi(table, "", &count);
-	fprintf(stderr, "%d nodes = %ld bytes.\n",
+	g_printerr("%d nodes = %ld bytes.\n",
 		count, (long) count * sizeof(struct _vte_table));
 }
 
@@ -646,7 +797,7 @@ escape(const char *p)
 	GString *ret;
 	int i;
 	guint8 check;
-	ret = g_string_new("");
+	ret = g_string_new(NULL);
 	for (i = 0; p[i] != '\0'; i++) {
 		tmp = NULL;
 		check = p[i];
@@ -703,11 +854,10 @@ print_array(GValueArray *array)
 			if (G_VALUE_HOLDS_POINTER(value)) {
 				printf("\"%ls\"",
 				       (wchar_t*) g_value_get_pointer(value));
-				g_free(g_value_get_pointer(value));
 			}
 		}
 		printf(")");
-		_vte_matcher_free_params_array(array);
+		/* _vte_matcher_free_params_array(array); */
 	}
 }
 
@@ -716,7 +866,7 @@ main(int argc, char **argv)
 {
 	struct _vte_table *table;
 	int i;
-	const char *patterns[] = {
+	const char *candidates[] = {
 		"ABCD",
 		"ABCDEF",
 		"]2;foo",
@@ -735,7 +885,7 @@ main(int argc, char **argv)
 	const char *result, *p;
 	const gunichar *consumed;
 	char *tmp;
-	gunichar *pattern;
+	gunichar *candidate;
 	GQuark quark;
 	GValueArray *array;
 	g_type_init();
@@ -748,25 +898,35 @@ main(int argc, char **argv)
 	_vte_table_add(table, "ACDEF%i%mJ", 10, "ACDEF%dJ", 0);
 	_vte_table_add(table, "[%mh", 5, "move-cursor", 0);
 	_vte_table_add(table, "[%d;%d;%dm", 11, "set-graphic-rendition", 0);
+	_vte_table_add(table, "[%dm", 5, "set-graphic-rendition", 0);
+	_vte_table_add(table, "[m", 3, "set-graphic-rendition", 0);
 	_vte_table_add(table, "]3;%s", 7, "set-icon-title", 0);
 	_vte_table_add(table, "]4;%s", 7, "set-window-title", 0);
 	printf("Table contents:\n");
 	_vte_table_print(table);
 	printf("\nTable matches:\n");
-	for (i = 0; i < G_N_ELEMENTS(patterns); i++) {
-		p = patterns[i];
-		pattern = make_wide(p);
+	for (i = 0; i < G_N_ELEMENTS(candidates); i++) {
+		p = candidates[i];
+		candidate = make_wide(p);
 		array = NULL;
-		_vte_table_match(table, pattern, strlen(p),
+		_vte_table_match(table, candidate, strlen(p),
 				 &result, &consumed, &quark, &array);
 		tmp = escape(p);
 		printf("`%s' => `%s'", tmp, (result ? result : "(NULL)"));
 		g_free(tmp);
 		print_array(array);
-		printf(" (%d chars)\n", consumed ? consumed - pattern : 0);
-		g_free(pattern);
+		printf(" (%d chars)\n", (int) (consumed ? consumed - candidate: 0));
+		g_free(candidate);
 	}
 	_vte_table_free(table);
 	return 0;
 }
 #endif
+
+const struct _vte_matcher_class _vte_matcher_table = {
+	(_vte_matcher_create_func)_vte_table_new,
+	(_vte_matcher_add_func)_vte_table_add,
+	(_vte_matcher_print_func)_vte_table_print,
+	(_vte_matcher_match_func)_vte_table_match,
+	(_vte_matcher_destroy_func)_vte_table_free
+};
